'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var inside = require('is-path-inside');
var upath = require('upath');
var chalk = require('chalk');
var superjson = require('superjson');
var consola = require('consola');
var path = require('path');
var pkgDir = require('pkg-dir');
var createJITI = require('jiti');
var _ = require('lodash');
var core = require('@midwayjs/core');
var tslib = require('tslib');
var decorator = require('@midwayjs/decorator');
var asynchronousLocalStorage = require('asynchronous-local-storage');
var fs = require('fs');
var globby = require('globby');
var staticCache = require('koa-static-cache');
var request_dist_midwayjsHooksCoreRequest = require('../request/dist/midwayjs-hooks-core-request.cjs.dev.js');
var esModuleLexer = require('es-module-lexer');
var art = require('art-template');
var prettier = require('prettier');
require('axios');

function _interopDefault (e) { return e && e.__esModule ? e : { 'default': e }; }

var inside__default = /*#__PURE__*/_interopDefault(inside);
var chalk__default = /*#__PURE__*/_interopDefault(chalk);
var superjson__default = /*#__PURE__*/_interopDefault(superjson);
var consola__default = /*#__PURE__*/_interopDefault(consola);
var path__default = /*#__PURE__*/_interopDefault(path);
var createJITI__default = /*#__PURE__*/_interopDefault(createJITI);
var ___default = /*#__PURE__*/_interopDefault(_);
var staticCache__default = /*#__PURE__*/_interopDefault(staticCache);
var art__default = /*#__PURE__*/_interopDefault(art);
var prettier__default = /*#__PURE__*/_interopDefault(prettier);

class ServerRouter {
  constructor(root, config, useSource = true) {
    this.root = void 0;
    this.config = void 0;
    this.routes = new Map();
    this.useSource = void 0;
    this.root = root;
    this.config = config;
    this.useSource = useSource;
  } // src/apis


  get source() {
    if (this.useSource) {
      return upath.join(this.root, this.config.source);
    }

    return upath.join(this.root, this.config.build.outDir);
  }

  getRouteConfig(file) {
    const {
      routes
    } = this.config;
    const index = routes.findIndex(route => {
      const dir = this.getApiDirectory(route.baseDir);
      return this.inside(file, dir);
    });

    if (!routes[index]) {
      return null;
    }

    return {
      underscore: false,
      ...routes[index]
    };
  } // src/apis/lambda


  getApiDirectory(baseDir) {
    return upath.join(this.source, baseDir);
  }

  isApiFile(file) {
    if (file.endsWith('.test.ts') || file.endsWith('.test.js')) {
      return false;
    }

    if (upath.extname(file) !== '.ts' && upath.extname(file) !== '.js') {
      return false;
    }

    const route = this.getRouteConfig(file);
    return !!route;
  }

  inside(child, parent) {
    return inside__default['default'](upath.toUnix(child), upath.toUnix(parent));
  }

  getBaseUrl(file) {
    const url = this.getHTTPPath(file, '', true);

    if (url === '/*') {
      return '/';
    }

    if (url.endsWith('/*')) {
      return url.slice(0, -2);
    }

    return url;
  }

  getHTTPPath(file, method, isExportDefault) {
    const {
      basePath,
      baseDir,
      underscore
    } = this.getRouteConfig(file);
    const lambdaDirectory = this.getApiDirectory(baseDir);
    const {
      isCatchAllRoutes,
      filename
    } = parseFilename(upath.basename(file, upath.extname(file)));
    const fileRoute = filename === 'index' ? '' : filename; // TODO remove underscore support in future

    const ApiMethodPrefix = '_';
    const methodPrefix = underscore ? ApiMethodPrefix : '';
    const func = isExportDefault ? '' : `${methodPrefix}${method}`;
    const api = upath.toUnix(upath.join(basePath,
    /**
     * /apis/lambda/index.ts -> ''
     * /apis/lambda/todo/index.ts -> 'todo'
     */
    upath.relative(lambdaDirectory, upath.dirname(file)),
    /**
     * index -> ''
     * demo -> '/demo'
     */
    fileRoute, // getTodoList -> _getTodoList
    func, isCatchAllRoutes ? '/*' : ''));
    /**
     * duplicate routes
     */

    const existPath = this.routes.get(api);

    if (existPath && existPath !== upath.toUnix(file)) {
      duplicateLogger(this.root, existPath, file, api);
    }

    this.routes.set(api, upath.toUnix(file));
    return api;
  }

}

function parseFilename(filename) {
  var _re$exec;

  const re = /\[\.{3}(.+)]/;
  const isCatchAllRoutes = re.test(filename);
  return {
    isCatchAllRoutes,
    filename: isCatchAllRoutes ? (_re$exec = re.exec(filename)) === null || _re$exec === void 0 ? void 0 : _re$exec[1] : filename
  };
}

function duplicateLogger(root, existPath, currentPath, api) {
  consola__default['default'].info('[ %s ] Duplicate routes detected. %s and %s both resolve to %s. Reference: %s', chalk__default['default'].yellow('warn'), chalk__default['default'].cyan(upath.relative(root, existPath)), chalk__default['default'].cyan(upath.relative(root, currentPath)), chalk__default['default'].cyan(api), 'https://www.yuque.com/midwayjs/faas/et7x4k');
}

function getProjectRoot(cwd) {
  return pkgDir.sync(cwd) || process.cwd();
}
function getConfig(cwd) {
  const root = getProjectRoot(cwd);
  const configs = {
    ts: path__default['default'].join(root, 'midway.config.ts'),
    js: path__default['default'].join(root, 'midway.config.js')
  };
  const userConfig = tryRequire(configs.ts) || tryRequire(configs.js);
  return ___default['default'].defaultsDeep({}, userConfig, {
    superjson: false,
    source: './src/apis',
    build: {
      viteOutDir: './build',
      outDir: './dist'
    },
    request: {
      client: '@midwayjs/hooks-core/request'
    }
  });
}
function defineConfig(config) {
  return config;
}

const tryRequire = id => {
  try {
    const jiti = createJITI__default['default']();
    const contents = jiti(id);
    if ('default' in contents) return contents.default;
    return contents;
  } catch {
    return undefined;
  }
};

function isDevelopment() {
  if (process.env.MIDWAY_TS_MODE === 'true' || process.env.NODE_ENV === 'test' || process.env.NODE_ENV === 'development') {
    return true;
  }

  return false;
}

/**
 * @private
 * private api, may change without notice.
 * Use asynchronous-local-storage due to serverless environment does not support node.js 12.17.0
 */
const als = {
  getStore(key) {
    return asynchronousLocalStorage.als.get(key);
  },

  run(ctx, callback) {
    return new Promise(resolve => {
      asynchronousLocalStorage.als.runWith(async () => {
        resolve(await callback());
      }, ctx);
    });
  }

};

class HooksDevComponent {
  constructor(config) {
    this.internal = void 0;
    this.runtime = void 0;
    this.router = void 0;
    this.root = void 0;
    this.container = void 0;

    this.useAsyncLocalStorage = async (ctx, next) => {
      const enableSuperjson = this.internal.superjson;
      await als.run({
        ctx
      }, async () => {
        try {
          await next();
        } catch (error) {
          if (enableSuperjson) {
            ctx.status = 500;
            ctx.body = superjson__default['default'].serialize(error);
          } else {
            throw error;
          }
        }
      });
    };

    this.runtime = config.runtime;
    this.internal = config.internal;
    this.router = config.router;
    this.root = config.root;
  }

  init() {
    const {
      routes
    } = this.internal;
    let count = 0;
    const totalCount = routes.reduce((totalCount, route) => {
      const dir = path.join(this.router.source, route.baseDir);
      const files = globby.sync([dir]).filter(file => this.router.isApiFile(file));
      return totalCount + files.length;
    }, 0);
    const configuration = core.createConfiguration({
      namespace: '@midwayjs/hooks',
      directoryResolveFilter: routes.map(route => {
        return {
          pattern: path.join(this.router.source, route.baseDir),
          ignoreRequire: true,
          filter: (_, file, container) => {
            if (!this.router.isApiFile(file)) {
              return;
            }

            this.container = container;

            const mod = require(file);

            this.createApi(mod, file);
            this.container.bindClass(mod, '', file);
            count++;

            if (count === totalCount) {
              this.afterCreate();
            }
          }
        };
      })
    });
    configuration.onReady((_, app) => {
      this.useGlobalMiddleware(app);
    }).onStop(_.noop);
    return {
      Configuration: configuration
    };
  }

  afterCreate() {}

  createApi(mod, file) {
    var _mod$config;

    const modMiddleware = (mod === null || mod === void 0 ? void 0 : (_mod$config = mod.config) === null || _mod$config === void 0 ? void 0 : _mod$config.middleware) || [];
    Object.keys(mod).filter(key => typeof mod[key] === 'function').forEach(key => {
      this.createApiFunction({
        fn: mod[key],
        file: file,
        isExportDefault: key === 'default',
        modMiddleware
      });
    });
  }

  createApiFunction(config) {
    const {
      fn,
      file,
      isExportDefault,
      modMiddleware
    } = config;
    const fnName = isExportDefault ? '$default' : fn.name;
    const id = this.getFunctionId(file, fnName, isExportDefault);
    const containerId = 'hooks::' + id;
    const httpPath = this.router.getHTTPPath(file, fnName, isExportDefault);
    const httpMethod = fn.length === 0 ? 'GET' : 'POST'; // Set param for unit testing

    fn._param = {
      url: httpPath,
      method: httpMethod,
      meta: {
        functionName: id
      }
    }; // Apply mod middleware

    fn.middleware = (fn.middleware || (fn.middleware = [])).concat(modMiddleware).map(this.useHooksMiddleware);
    this.registerApiFunction({
      containerId,
      httpMethod,
      httpPath,
      fn
    });
  }

  getFunctionId(file, functionName, isExportDefault) {
    const relativePath = upath.relative(this.router.source, file); // src/apis/lambda/index.ts -> apis-lambda-index

    const id = _.kebabCase(upath.removeExt(relativePath, upath.extname(relativePath)));
    const name = [id, isExportDefault ? '' : `-${functionName}`].join('');
    return name.toLowerCase();
  }

  registerApiFunction(config) {
    const {
      containerId,
      httpMethod,
      httpPath,
      fn
    } = config;
    const Method = httpMethod === 'GET' ? decorator.Get : decorator.Post;
    const enableSuperjson = this.internal.superjson;
    let FunctionContainer = class FunctionContainer {
      constructor() {
        this.ctx = void 0;
      }

      async handler() {
        var _this$ctx$request, _this$ctx$request$bod;

        let args = ((_this$ctx$request = this.ctx.request) === null || _this$ctx$request === void 0 ? void 0 : (_this$ctx$request$bod = _this$ctx$request.body) === null || _this$ctx$request$bod === void 0 ? void 0 : _this$ctx$request$bod.args) || [];

        if (typeof args === 'string') {
          args = JSON.parse(args);
        }

        const response = await fn(...args);

        if (enableSuperjson) {
          return superjson__default['default'].serialize(response);
        }

        return response;
      }

    };

    tslib.__decorate([decorator.Inject()], FunctionContainer.prototype, 'ctx', void 0);

    tslib.__decorate([Method(httpPath, {
      middleware: fn.middleware
    })], FunctionContainer.prototype, 'handler', null);

    FunctionContainer = tslib.__decorate([decorator.Provide(containerId), decorator.Controller('/')], FunctionContainer);
    this.container.bind(containerId, FunctionContainer);
  }

  useGlobalMiddleware(app) {
    app.use(this.useAsyncLocalStorage); // Apply global middleware from config

    if (Array.isArray(this.runtime.middleware)) {
      this.runtime.middleware.forEach(middleware => app.use(this.useHooksMiddleware(middleware)));
    }
  }

  useHooksMiddleware(fn) {
    return (...args) => {
      /**
       * Hooks middleware
       * const middleware = (next) => { const ctx = useContext() }
       */
      if (fn.length === 1) {
        const next = args[args.length - 1];
        return fn(next);
      }

      return fn(...args);
    };
  }

  isFullStackProject() {
    return [// Vite
    'vite.config.ts', 'vite.config.js', // build-scritps
    'build.json', 'build.js'].some(config => fs.existsSync(path.join(this.root, config)));
  }

}

class HooksProductionComponent extends HooksDevComponent {
  useGlobalMiddleware(app) {
    super.useGlobalMiddleware(app);

    if (!this.isFullStackProject()) {
      return;
    }

    if (this.existRootPath) {
      return;
    }

    const baseDir = app.getBaseDir();
    const dist = upath.join(baseDir, '..', this.internal.build.viteOutDir);
    const mw = staticCache__default['default']({
      dir: dist,
      dynamic: true,
      alias: {
        '/': 'index.html'
      },
      buffer: true,
      gzip: true
    });
    app.use(mw);
  }

  afterCreate() {
    if (!this.isFullStackProject()) {
      return;
    }

    if (this.existRootPath) {
      consola__default['default'].debug('The route `/` or `/*` is registered, you need to host the front-end page manually');
      return;
    }

    this.registerApiFunction({
      containerId: 'hooks:host',
      httpMethod: 'GET',
      httpPath: '/*',
      fn: async () => {}
    });
  }

  get existRootPath() {
    const {
      routes
    } = this.router;
    return routes.has('/') || routes.has('/*');
  }

}

/**
 * Create hooks component
 */

const hooks = (runtime = {}) => {
  const root = getProjectRoot();
  const internal = getConfig();
  const router = new ServerRouter(root, internal, isDevelopment());
  const Component = isDevelopment() ? HooksDevComponent : HooksProductionComponent;
  const cmp = new Component({
    runtime,
    internal,
    router,
    root
  });
  return cmp.init();
};

// TODO Refactor SDK generator to support custom request sdk
async function generate(baseUrl, code, superjson, sdk) {
  await esModuleLexer.init;
  const [, exports] = esModuleLexer.parse(code);

  if (exports.length === 0) {
    return null;
  }

  const funcs = exports.map(name => {
    return {
      url: getUrl(baseUrl, name),
      meta: {
        superjson
      },
      isExportDefault: name === 'default',
      name: name === 'default' ? '$default' : name
    };
  });
  const templates = {
    dev: path.resolve(__dirname, '../../template/request.art'),
    prod: path.resolve(__dirname, '../template/request.art')
  };
  const templatePath = fs.existsSync(templates.dev) ? templates.dev : templates.prod;
  const template = fs.readFileSync(templatePath, {
    encoding: 'utf-8'
  });
  return prettier__default['default'].format(art__default['default'].render(template, {
    funcs,
    SDK: sdk
  }), {
    semi: true,
    singleQuote: true,
    parser: 'babel'
  });
}

function getUrl(baseUrl, name) {
  if (name === 'default') {
    return baseUrl;
  }

  if (baseUrl.endsWith('/')) {
    return `${baseUrl}${name}`;
  }

  return `${baseUrl}/${name}`;
}

art__default['default'].defaults.htmlMinifierOptions = {
  collapseWhitespace: false,
  minifyCSS: false,
  minifyJS: false,
  ignoreCustomFragments: []
};

art__default['default'].defaults.imports.stringify = json => JSON.stringify(json, null, 2);

art__default['default'].defaults.escape = false;

Object.defineProperty(exports, 'superjson', {
  enumerable: true,
  get: function () {
    return superjson__default['default'];
  }
});
Object.defineProperty(exports, 'consola', {
  enumerable: true,
  get: function () {
    return consola__default['default'];
  }
});
exports.defaults = request_dist_midwayjsHooksCoreRequest.defaults;
exports.request = request_dist_midwayjsHooksCoreRequest.request;
exports.ServerRouter = ServerRouter;
exports.als = als;
exports.defineConfig = defineConfig;
exports.generate = generate;
exports.getConfig = getConfig;
exports.getProjectRoot = getProjectRoot;
exports.hooks = hooks;
exports.isDevelopment = isDevelopment;
